@@ -1,5 +1,5 @@
 #                                                         -*- Perl -*-
-# Copyright (c) 1999, 2000  Motoyuki Kasahara
+# Copyright (c) 1999, 2000, 2001  Motoyuki Kasahara
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@ use FileHandle;
 use FreePWING::Reference;
 use FreePWING::Tag;
 use FreePWING::RefUserChar;
+use FreePWING::RefSound;
 use FreePWING::CharConv;
 use strict;
 use integer;
@@ -47,6 +48,11 @@ $block_length = 2048;
 #
 $max_indent_level = 6;
 
+#
+# Хåե
+#
+my $max_buffer_size = 64 * 1024;
+
 #
 # :
 #	new()
@@ -66,6 +72,9 @@ sub new {
 	# ʸե̾
 	'file_name' => '',
 
+	# ϥХåե
+	'buffer' => '',
+
 	# ƥȿ
 	'context_count' => 0,
 
@@ -105,6 +114,9 @@ sub new {
 	# ѳ
 	'full_user_characters' => FreePWING::RefUserChar->new(),
 
+	#  (satomii)
+	'sounds' => FreePWING::RefSound->new(),
+
 	# 顼å
 	'error_message' => '',
     };
@@ -210,6 +222,14 @@ sub close {
 	return 0;
     }
 	
+
+    #
+    # ХåեƤե˽񤭤
+    #
+    if (!$self->flush_buffer()) {
+	return 0;
+    }
+
     #
     # եĤ롣
     #
@@ -290,6 +310,32 @@ sub set_full_user_characters_in_file {
     return 1;
 }
 
+#
+# :
+#	set_sounds_in_file(file_name)
+#           file_name
+#		ե̾
+# ᥽åɤζʬ:
+#	public 󥹥󥹥᥽åɡ
+# :
+#	եɤ߹ࡣ
+# :
+#	 1 ֤Ԥ 0 ֤
+#
+# (satomii)
+#
+sub set_sounds_in_file {
+    my $self = shift;
+    my ($file_name) = @ARG;
+
+    if (!$self->{'sounds'}->add_sounds_in_file($file_name)) {
+	$self->{'error_message'} = $self->{'sounds'}->error_message();
+	$self->close_internal();
+	return 0;
+    }
+    return 1;
+}
+
 #
 # :
 #	write_data(data)
@@ -306,14 +352,44 @@ sub write_data {
     my $self = shift;
     my ($data) = @ARG;
 
-    if (!$self->{'handle'}->print($data)) {
+    $self->{'buffer'} .= $data;
+
+    if ((length($self->{'buffer'}) > $max_buffer_size)
+	&& !$self->flush_buffer()) {
+      return 0;
+    }
+
+    $self->{'position'} += length($data);
+    return 1;
+}
+
+#
+# :
+#	flush_buffer()
+# ᥽åɤζʬ:
+# 	private 󥹥󥹥᥽åɡ
+# :
+# 	ХåեƤե˽񤭹ߡХåեˤޤ
+# :
+#	 1 ֤Ԥ 0 ֤
+#
+sub flush_buffer {
+    my $self = shift;
+
+    if (!length($self->{'buffer'})) {
+      return 1;
+    }
+
+    if (!$self->{'handle'}->print($self->{'buffer'})) {
 	$self->{'error_message'} =
 	    "failed to write the file, $ERRNO: " . $self->{'file_name'};
+	$self->{'buffer'} = '';
 	$self->close_internal();
 	return 0;
     }
 
-    $self->{'position'} += length($data);
+    $self->{'buffer'} = '';
+    return 1;
 }
 
 #
@@ -553,87 +629,67 @@ sub add_text {
     my ($text) = @ARG;
 
     #
-    # ʸ˽
+    #  ("\r", "\n") ̵롣
     #
-    my @unpacked_text = unpack('C*', $text);
-    my $i = 0;
-    while ($i < scalar(@unpacked_text)) {
-	my $c0 = $unpacked_text[$i];
-
+    $text =~ s/[\r\n]//g;
 	#
 	# ʿ ("\t") ϡޤѴ
 	# 
-	if ($c0 == 0x09) {
-	    $c0 = 0x20;
-	}
+    $text =~ s/\t/ /g;
 
-	if ($c0 == 0x0a || $c0 == 0x0d) {
-	    #
-	    #  ("\r", "\n") ̵롣
-	    #
-	    $i++;
-	    next;
-	} elsif (0x20 <= $c0 && $c0 <= 0x7e) {
+    my $word;
+    for (;;) {
+	if ($text =~ s/^([\x20-\x7f]+)//) {
 	    #
 	    # US-ASCII ȾѰ
 	    #
+	    $word = $1;
 	    if ($self->{'modifier_stack'}->[-1] ne 'half-width'
 		&& !$self->add_half_width_start()) {
 		return 0;
 	    }
-	    if (!$self->write_data($ascii_to_jisx0208_table->[$c0 - 0x20])) {
+	    $word =~ s/(.)/$ascii_to_jisx0208_table->[unpack("C",$1)-0x20]/g;
+	    if (!$self->write_data($word)) {
 		return 0;
 	    }
-	    $i++;
-	} elsif (0xa1 <= $c0 && $c0 <= 0xfe) {
+	} elsif ($text =~ s/^(([\xa1-\xfe]{2})+)//) {
 	    #
 	    # JIS X 0208 ʸϤΤޤ޵Ͽ
 	    #
-	    my $c1 = $unpacked_text[$i + 1];
-	    if (!defined($c1) || $c1 < 0xa1 || 0xfe < $c1) {
-		$self->{'error_message'} =
-		    sprintf("invalid character: \\x%02x", $c0);
-		$self->close_internal();
-		return 0;
-	    }
+	    $word = $1;
 	    if ($self->{'modifier_stack'}->[-1] eq 'half-width'
 		&& !$self->add_half_width_end()) {
 		return 0;
 	    }
-	    if (!$self->write_data(pack("CC", $c0 & 0x7f, $c1 & 0x7f))) {
+	    $word =~ tr/\xa1-\xfe/\x21-\x7e/;
+	    if (!$self->write_data($word)) {
 		return 0;
 	    }
-	    $i += 2;
-	} elsif ($c0 == 0x8e) {
+	} elsif ($text =~ s/^((\x8e[\xa1-\xfe])+)//) {
 	    #
 	    # SS2 Ѥ JIS X 0201 ʤ JIS X 0208 ʤѴ
 	    #
-	    my $c1 = $unpacked_text[$i + 1];
-	    if (!defined($c1) || $c1 < 0xa1 || 0xfe < $c1) {
-		$self->{'error_message'} =
-		    sprintf("invalid character: \\x%02x", $c0);
-		$self->close_internal();
-		return 0;
-	    }
+	    $word = $1;
 	    if ($self->{'modifier_stack'}->[-1] eq 'half-width'
 		&& !$self->add_half_width_end()) {
 		return 0;
 	    }
-	    if (!$self->write_data($jisx0201_to_jisx0208_table
-				   ->[$c1 - 0xa0])) {
+	    $word =~ s/\x8e(.)/$jisx0201_to_jisx0208_table->[unpack("C",$1)-0xa0]/g;
+	    if (!$self->write_data($word)) {
 		return 0;
 	    }
-	    $i += 2;
 	} else {
-	    # ¾ʸϥ顼
+	    if(length($text)) {
 	    $self->{'error_message'} =
-		sprintf("invalid character: \\x%02x", $c0);
+		    sprintf("invalid character: \\x%02x", unpack("C", $text));
 	    $self->close_internal();
 	    return 0;
-	}
-    }
+	    } else {
     $self->{'new_entry_skip_flag'} = 0;
     return 1;
+	    }
+	}
+    }
 }
 
 #
@@ -745,7 +801,8 @@ sub add_full_user_character {
 sub add_entry_tag {
     my $self = shift;
     my ($tag) = @ARG;
-    
+ 
+    $tag =~ s/^/text:/;
     return $self->{'tag'}->add_entry($tag, $self->{'entry_position'});
 }
 
@@ -765,6 +822,7 @@ sub add_tag {
     my $self = shift;
     my ($tag) = @ARG;
     
+    $tag =~ s/^/text:/;
     return $self->{'tag'}->add_entry($tag, $self->{'position'});
 }
 
@@ -875,13 +933,14 @@ sub add_reference_start {
 }
 
 #
-# ̹ܻȽλ
+# ܻȽλ
 # ΰ֤򼨤̾ǻꡣ
 #
 sub add_reference_end {
     my $self = shift;
     my ($target_position) = @ARG;
 
+    $target_position =~ s/^/text:/;
     if (!$self->add_modifier_end('reference', pack('nNn', 0x1f62, 0, 0))) {
 	return 0;
     }
@@ -894,6 +953,127 @@ sub add_reference_end {
     return 1;
 }
 
+#
+# 顼ǻ (BMP )
+# 顼̾ǻꡣ
+#
+sub add_color_graphic_start {
+    my $self = shift;
+    my ($picture_name) = @ARG;
+
+    $picture_name =~ s/^/cgraph:/;
+    if (!$self->add_modifier_start('color-graphic',
+				   pack('n3N3n', 0x1f4d, 0x0009, 0x0001,
+					0, 0, 0, 0))) {
+	return 0;
+    }
+    if (!$self->{'reference'}
+	->add_tag_entry($self->{'position'} - 6, $picture_name)) {
+	$self->{'error_message'} = $self->{'reference'}->error_message();
+	$self->close_internal();
+	return 0;
+    }
+    return 1;
+}
+
+#
+# 顼ǻ (JPG )
+# 顼̾ǻꡣ
+#
+sub add_jpeg_graphic_start {
+    my $self = shift;
+    my ($picture_name) = @ARG;
+
+    $picture_name =~ s/^/cgraph:/;
+    if (!$self->add_modifier_start('color-graphic',
+				   pack('n3N3n', 0x1f4d, 0x1209, 0x0001,
+					0, 0, 0, 0))) {
+	return 0;
+    }
+
+    if (!$self->{'reference'}
+	->add_tag_entry($self->{'position'} - 6, $picture_name)) {
+	$self->{'error_message'} = $self->{'reference'}->error_message();
+	$self->close_internal();
+	return 0;
+    }
+    return 1;
+}
+
+#
+# 顼ǻȽλ (BMP )
+#
+sub add_color_graphic_end {
+    my $self = shift;
+
+    if (!$self->add_modifier_end('color-graphic', pack('n', 0x1f6d))) {
+	return 0;
+    }
+    return 1;
+}
+
+#
+# 顼ǻȽλ (JPG )
+#
+sub add_jpeg_graphic_end {
+    my $self = shift;
+
+    if (!$self->add_modifier_end('color-graphic', pack('n', 0x1f6d))) {
+	return 0;
+    }
+    return 1;
+}
+
+#
+# 
+# ǡ̾ǻꡣ
+#
+sub add_sound_start {
+    my $self = shift;
+    my ($sound_name) = @ARG;
+
+    #  FreePWING::RefSound ͳǼ롣(satomii)
+    if (!$self->add_modifier_start(
+	     'sound',
+	     pack('nC2nNnNn', 0x1f4a, 0,
+		  $self->{'sounds'}->sound_type($sound_name),
+		  $self->{'sounds'}->sound_format($sound_name),
+		  0, 0, 0, 0))) {
+	return 0;
+    }
+
+    # if (!$self->add_modifier_start('sound',
+    #				   pack('n3N3', 0x1f4a, 0, 0x0012, 0, 0, 0))) {
+    #	return 0;
+    # }
+    if (!$self->{'reference'}
+	->add_tag_entry($self->{'position'} - 12, "sound:$sound_name")) {
+	$self->{'error_message'} = $self->{'reference'}->error_message();
+	$self->close_internal();
+	return 0;
+    }
+    if (!$self->{'reference'}
+	->add_tag_entry($self->{'position'} - 6, "sound-end:$sound_name")) {
+	$self->{'error_message'} = $self->{'reference'}->error_message();
+	$self->close_internal();
+	return 0;
+    }
+    return 1;
+}
+
+#
+# Ƚλ
+# ΰ֤򼨤̾ǻꡣ
+#
+sub add_sound_end {
+    my $self = shift;
+
+    if (!$self->add_modifier_end('sound', pack('n', 0x1f6a))) {
+	return 0;
+    }
+    return 1;
+}
+
 #
 # Ⱦѳϻ
 #
